@@ -63,7 +63,9 @@ def setup(**arg_dict):
         'install_kcm' : InstallKControlModule,
         'build_html' : BuildDocbookHTML,
         'install_html' : InstallDocbookHTML,
-        'install_lib' : InstallLibWithRoot }
+        'install_lib' : InstallLibWithRoot,
+        'build_kioslave' : BuildKioslave,
+        'install_kioslave' : InstallKioslave}
 
     for key in kdecmdclass.iterkeys():
         cmdclass.setdefault(key,kdecmdclass[key])
@@ -78,6 +80,7 @@ class KDEDistribution(Distribution):
         self.min_kde_version = None
         self.min_qt_version = None
         self.kcontrol_modules = None
+        self.kioslaves = None
         self.executable_links = None
         self.docbooks = None
         self.application_data = None
@@ -105,6 +108,11 @@ def has_application_data(self):
         return 0
     return  len(self.distribution.application_data)!=0
 
+def has_kioslaves(self):
+    if self.distribution.kioslaves is None:
+        return 0
+    return  len(self.distribution.kioslaves)!=0
+
 ###########################################################################
 # Our slightly extended build command. This also does kcontrol modules.
 class BuildKDE(build):
@@ -115,12 +123,14 @@ class BuildKDE(build):
 
     sub_commands = build.sub_commands[:]
     sub_commands.append( ('build_kcm',has_kcms) )
+    sub_commands.append( ('build_kioslave',has_kioslaves) )
     sub_commands.append( ('build_html',has_docbook_html) )
     sub_commands.append( ('build_messages',has_messages) )
     
     def __init__(self,dist):
         build.__init__(self,dist)
         self.has_kcms = has_kcms
+        self.has_kioslaves = has_kioslaves
         self.has_docbook_html = has_docbook_html
         self.has_messages = has_messages
 
@@ -163,6 +173,8 @@ class InstallKDE(install):
     user_options.append( ('install-cmd=',None,"Command to use to install the files") )
     user_options.append( ('install-xdg-apps=',None,"directory for XDG app files") )
     user_options.append( ('install-kcm=',None,"directory for kcm library files") )
+    user_options.append( ('install-kioslave',None,"directory for kioslave library files") )
+    user_options.append( ('install-protocol',None,"directory for kioslave protocol files") )
     
     sub_commands = install.sub_commands[:]
     sub_commands.insert(0, ('checkpykde',None) )
@@ -171,6 +183,7 @@ class InstallKDE(install):
     sub_commands.append( ('install_messages',has_messages) )
     sub_commands.append( ('install_html',has_docbook_html) )
     sub_commands.append( ('install_kcm',has_kcms) )
+    sub_commands.append( ('install_kioslave',has_kioslaves) )
     sub_commands.append( ('install_application_data',has_application_data) )
         
     def initialize_options(self):
@@ -180,6 +193,8 @@ class InstallKDE(install):
         self.install_cmd = None
         self.install_xdg_apps = None
         self.install_kcm = None
+        self.install_kioslave = None
+        self.install_protocol = None
         self.install_application_data = None
         install.initialize_options(self)
         
@@ -244,6 +259,16 @@ class InstallKDE(install):
             self.install_kcm = os.path.join(ask_kde_config('--install lib --expandvars').strip(),'kde3')
             self.announce("  ...kcm library directory is %s" % self.install_kcm)
     
+        if self.install_kioslave is None:
+            self.announce("Detecting kioslave library  directory...")
+            self.install_kioslave = os.path.join(ask_kde_config('--install lib --expandvars').strip(),'kde3')
+            self.announce("  ...kioslave library directory is %s" % self.install_kioslave)
+    
+        if self.install_protocol is None:
+            self.announce("Detecting kioslave protocol directory...")
+            self.install_protocol = ask_kde_config('--install services --expandvars').strip()
+            self.announce("  ...kioslave protocol directory is %s" % self.install_protocol)
+    
         install.finalize_options(self)
         
         if self.root is not None:
@@ -297,7 +322,7 @@ class InstallApplicationDataAndLinks(install_data):
                     out = self.copy_tree(f,os.path.join(self.install_dir,f))
                     self.outfiles.extend(out)
                 else:
-                    self.warn("Setup script can't find file or directory %f needed for installation." % f)
+                    self.warn("Setup script can't find file or directory %s needed for installation." % f)
             else:
                 # it's a tuple with path to install to and a list of files
                 dir = convert_path(f[0])
@@ -702,7 +727,8 @@ class BuildKControlModule(Command):
 
             # Couldn't get it to pass without this ...
             cmdlist.append("--mode=compile")
-
+            cmdlist.append("--tag=CXX")
+            
             # Find the compiler flags and options
             # CXX is empty on some Systems, let's do it 'the hard way'.
             # FIXME :: get CXX from make.conf for Gentoo.
@@ -1157,7 +1183,7 @@ class InstallDocbookHTML(Command):
             for source_file in glob.glob(os.path.join(input_dir,'*.html')):
                 htmlfile = os.path.basename(source_file)
                 dest_file = os.path.join(dest_path,htmlfile)
-                self.copy_file(source_file,dest_file)
+                self.copy_file(source_file, dest_file)
                 self.outfiles.append(dest_file)
                 counter += 1
             
@@ -1166,7 +1192,7 @@ class InstallDocbookHTML(Command):
                 for file in extra_files:
                     source_file = os.path.join(input_dir,file)
                     dest_file = os.path.join(self.install_dir, language_code, self.distribution.metadata.name,file)
-                    self.copy_file(source_file,dest_file)
+                    self.copy_file(source_file, dest_file)
                     self.outfiles.append(dest_file)
                     counter += 1
         
@@ -1368,7 +1394,547 @@ class InstallI18NMessages(Command):
 
     def mkpath(self, name, mode=0777):
         return dir_util.mkpath(name, mode, dry_run=self.dry_run)
-       
+
+ 
+###########################################################################
+class BuildKioslave(Command):
+    description = "Build Kioslaves"
+
+    user_options = [('no-kioslave',None,"Don't build kioslaves"),
+                    ('build-dir=','b', "build directory (where to install from)"),
+                    ('python-dir=',None,'Directory containing the Python installation'),
+                    ('python-inc-dir=',None,'Directory containing C Header files for Python'),
+                    ('kde-inc-dir=',None,'Directory containing C++ header files for KDE'),
+                    ('kde-lib-dir=',None,'Directory containing library files for KDE'),
+                    ('kde-kioslave-lib-dir=',None,'Directory for KDE kioslave library files'),
+                    ('kde-protocol-dir=',None,'Directory for KDE kioslave protocol files'),
+                    ('qt-inc-dir=',None,'Directory containing C++ header files for Qt'),
+                    ('qt-lib-dir=',None,'Directory containing library files for Qt'),
+                    ('sip-dir=',None,'Directory containing the sip library files'),
+                    ('clib=',None,'gcc library and path'),
+                    ('pyqt-dir=',None,'PyQt module directory'),
+                    ('pykde-dir=',None,'PyKDE module directory'),
+                    ('data-dir=',None,'installation directory for data (script) files')]
+    
+    def initialize_options(self):
+        self.no_kioslave = 0
+        self.build_dir = None
+        self.python_inc_dir = None
+        self.python_dir = None
+        self.kde_inc_dir = None
+        self.kde_lib_dir = None
+        self.kde_kioslave_lib_dir = None
+        self.kde_protocol_dir = None
+        self.qt_inc_dir = None
+        self.qt_lib_dir = None
+        self.sip_dir = None
+        self.clib = None
+        self.pyqt_dir = None
+        self.pykde_dir = None
+        self.data_dir = None
+        
+    def finalize_options(self):
+        if self.no_kioslave==0:
+            self.set_undefined_options('install', ('build_base', 'build_dir'),('install_application_data','data_dir'))
+        
+            install = self.get_finalized_command('install')
+            self.install_prefix = install.prefix
+    
+            # KDE inc dir: find it!
+            if self.kde_inc_dir is None:
+                canidatepaths = []
+                kdedir = os.getenv("KDEDIR")
+                if kdedir!=None:
+                    canidatepaths.append(os.path.join(kdedir,"include"))
+                canidatepaths.append(os.path.join(install.prefix,"include"))
+                canidatepaths.append('/opt/kde3/include')
+                canidatepaths.append('/opt/kde/include')
+                self.kde_inc_dir = FindFileInPaths('kapplication.h',canidatepaths)
+            if self.kde_inc_dir is None:
+                raise SystemExit, "Failed to find the KDE header file directory."
+            if FindFileInPaths('kapplication.h',[self.kde_inc_dir]) is None:
+                raise SystemExit, "Failed to find KDE header files in: %s" % self.kde_inc_dir
+            self.announce("Using %s for KDE header files" % self.kde_inc_dir)
+    
+            # KDE lib dir
+            if self.kde_lib_dir is None:
+                self.kde_lib_dir = os.path.join(install.prefix,"lib")
+            self.announce("Using %s for KDE library files" % self.kde_lib_dir)
+    
+            # KDE kioslave lib dir
+            if self.kde_kioslave_lib_dir is None:
+                self.kde_kioslave_lib_dir = os.path.join(self.kde_lib_dir,"kde3")
+            if FindFileInPaths('kio_*.so',[self.kde_kioslave_lib_dir]) is None:
+                raise SystemExit, "Failed to find KDE Kioslave library files in: %s" % self.kde_kioslave_lib_dir
+            self.announce("Using %s for KDE Kioslave library files" % self.kde_kioslave_lib_dir)
+            
+            # Qt inc dir
+            if self.qt_inc_dir is None:
+                canidatepaths = []
+                qtdir = os.getenv("QTDIR")
+                if qtdir != None:
+                    canidatepaths.append(os.path.join(qtdir,"include"))
+                canidatepaths.append(os.path.join(install.prefix,"lib/qt3/include"))
+                canidatepaths.append(os.path.join(install.prefix,"lib/qt3/include"))
+                canidatepaths.append(os.path.join(install.prefix,"include"))
+                canidatepaths.append("/opt/qt3/include")
+                canidatepaths.append("/opt/qt/include")
+                canidatepaths.append("/opt/qt/lib/include")
+                canidatepaths.append("/opt/qt3/lib/include")
+                self.qt_inc_dir = FindFileInPaths('qstring.h',canidatepaths)
+            if self.qt_inc_dir is None:
+                raise SystemExit,"Failed to find the Qt header file directory"
+            if FindFileInPaths('qstring.h',[self.qt_inc_dir]) is None:
+                raise SystemExit, "Failed to find Qt header files in: %s" % self.qt_inc_dir
+            self.announce("Using %s for Qt header files" % self.qt_inc_dir)
+    
+            # Qt lib dir
+            if self.qt_lib_dir is None:
+                canidatepaths = []
+                qtdir = os.getenv("QTDIR")
+                if qtdir != None:
+                    canidatepaths.append(os.path.join(qtdir,get_libdir_name()))
+                canidatepaths.append(os.path.join(install.prefix,"lib/qt3/"+get_libdir_name()))
+                canidatepaths.append(os.path.join(install.prefix,"lib/qt3/"+get_libdir_name()))
+                canidatepaths.append(os.path.join(install.prefix,get_libdir_name()))
+                canidatepaths.append("/opt/qt3/"+get_libdir_name())
+                canidatepaths.append("/opt/qt/"+get_libdir_name())
+                canidatepaths.append("/opt/qt/lib/"+get_libdir_name())
+                canidatepaths.append("/opt/qt3/lib/"+get_libdir_name())
+                self.qt_lib_dir = FindFileInPaths('libqt*',canidatepaths)
+            if self.qt_lib_dir is None:
+                raise SystemExit, "Failed to find Qt library files"
+            self.announce("Using %s for Qt library files" % self.qt_lib_dir)
+                    
+            # Python dir
+            if self.python_dir is None:
+                self.python_dir = os.path.split(sysconfig.get_config_var("LIBP"))[0]
+            self.announce("Using %s for the python directory" % self.python_dir)
+            
+            # Python include dir.
+            if self.python_inc_dir is None:
+                # Find the Python include directory.
+                self.python_inc_dir = sysconfig.get_config_var("INCLUDEPY")
+            self.announce("Using %s for Python header files" % self.python_inc_dir)
+    
+            # PyQt dir
+            if self.pyqt_dir is None:
+                self.pyqt_dir = sysconfig.get_python_lib()
+            if (FindFileInPaths("libqtcmodule*",[self.pyqt_dir]) is None) and (FindFileInPaths("qt*",[self.pyqt_dir]) is None):
+                raise SystemExit, "Failed to find the PyQt directory: %s" % self.pyqt_dir
+            self.announce("Using %s for PyQt modules" % self.pyqt_dir)
+    
+            # PyKDE dir
+            if self.pykde_dir is None:
+                self.pykde_dir = sysconfig.get_python_lib()
+            if (FindFileInPaths("libkdecorecmodule*",[self.pykde_dir]) is None) and (FindFileInPaths("kdecore*",[self.pykde_dir]) is None):
+                raise SystemExit, "Failed to find the PyKDE directory: %s" % self.pykde_dir
+            self.announce("Using %s for PyKDE modules" % self.pykde_dir)
+            
+            # Sip dir
+            if self.sip_dir is None:
+                self.sip_dir = sysconfig.get_python_lib()
+            if (FindFileInPaths("libsip*", [self.sip_dir]) is None) and (FindFileInPaths("sip*", [self.sip_dir]) is None):
+                raise SystemExit, "Failed to find libsip files in directory: %s" % self.sip_dir
+            self.announce("Using %s for libsip files" % self.sip_dir)
+                
+            # Find the C library (libgcc, libgcc_s or some other variation).
+            if self.clib is None:
+                canidatepaths = ["/usr/"+get_libdir_name(), "/usr/local/"+get_libdir_name() ]
+                self.clib = FindFileInPaths("libgcc*.so",canidatepaths)
+                if self.clib!=None:
+                    self.clib = glob.glob(os.path.join(self.clib,'libgcc*.so'))[0]
+                else:
+                    self.clib = FindFileInPaths("libgcc*.a",canidatepaths)
+                    if self.clib!=None:
+                        self.clib = glob.glob(os.path.join(self.clib,'libgcc*.a'))[0]
+            if self.clib is None:
+                raise SystemExit, "Failed to find a suitable libgcc library"
+            self.announce("Using %s for clib" % self.clib)
+    
+            # Make a list of places to look for python .so modules
+            self.python_sub_dirs = sysconfig.get_config_var("LIBSUBDIRS").split()
+            base = sysconfig.get_config_var("LIBP")
+            self.python_sub_dirs = [ os.path.join(base,item) for item in self.python_sub_dirs ]
+            self.python_sub_dirs.append(base)
+
+    def get_command_name(self):
+        return 'build_kioslave'
+
+    def run(self):
+        if self.no_kioslave:
+            self.announce("Skipping KIO Slaves")
+            return
+    
+        if not os.path.isdir(self.build_dir):
+            os.mkdir(self.build_dir)
+        
+        for moduletuple in self.distribution.kioslaves:
+            self.announce("Building KIO Slave from protocol file %s." % moduletuple[0])
+            
+            protocolfilename = moduletuple[0]
+            
+            # Read the protocol file
+            libraryname = None
+            try:
+                fhandle = open(protocolfilename,'r')
+                for line in fhandle.readlines():
+                    parts = line.strip().split('=')
+                    try:
+                        if parts[0]=="exec":
+                            libraryname = parts[1]
+                    except IndexError:
+                        pass
+                fhandle.close()
+            except IOError:
+                raise SystemExit, "Failed to find kioslave protocol file: %s" % moduletuple[0]
+                
+            # Sanity check.
+            if libraryname is None:
+                raise SystemExit, "Failed to find library name (Was there a exec entry in the protocol file?)"
+            
+            modulename = os.path.basename(moduletuple[1])
+            if modulename.endswith('.py'):
+                modulename = modulename[:-3]
+            
+            stub_cpp_name = libraryname+'.cpp'
+            stub_so_name = libraryname+'.so'
+            stub_la_name = libraryname+'.la'
+            python_version = '%i.%i' % (sys.version_info[0],sys.version_info[1])
+            
+            # Build the 'stub' code.
+            cppcode = self.cpptemplate % {"moduledir": self.data_dir,
+                                            "modulename": modulename,
+                                            "python_version": python_version}
+            
+            # Put it on disk.
+            cppfile = os.path.join(os.path.dirname(moduletuple[0]),stub_cpp_name)
+            try:
+                fhandle = open(cppfile,'w')
+                fhandle.write(cppcode)
+                fhandle.close()
+            except IOError:
+                raise SystemExit, "Could not write the C++ stub: %s" % cppfile
+                
+            # Compile the stub library.
+            cmdlist = ['libtool']
+
+            # Couldn't get it to pass without this ...
+            cmdlist.append("--mode=compile")
+            cmdlist.append("--tag=CXX")
+            
+            # Find the compiler flags and options
+            # CXX is empty on some Systems, let's do it 'the hard way'.
+            # FIXME :: get CXX from make.conf for Gentoo.
+            if len(sysconfig.get_config_var("CXX").split()) >= 2:
+                cmdlist.extend(sysconfig.get_config_var("CXX").split())
+            else:
+                cmdlist.extend(['g++', '-pthread'])
+
+            #cmdlist.extend(sysconfig.get_config_var("CXX").split())
+            
+            # cc_flags
+            cmdlist.append("-c")
+            cmdlist.append("-g")
+            
+            # The 4 is randomly chosen!
+            # FIXME :: get CFLAGS from make.conf for Gentoo.
+            if len(sysconfig.get_config_var("CFLAGS").split()) >=4:
+                cmdlist.extend(sysconfig.get_config_var("CFLAGS").split())
+            else:
+                # On Gentoo systems, CFLAGS are not in the environment.
+                raw = os.popen('emerge info 2> /dev/null|grep CFLAGS')
+                lines = raw.readlines()
+                if len(lines):
+                    cflags = lines[0].split('"')[1].split()
+                    print "Got CFLAGS from emerge info."
+                    cmdlist.extend(cflags)
+                else:
+                    # Still no CFLAGS found, use these ...
+                    cmdlist.extend(['-fno-strict-aliasing', '-DNDEBUG', '-g', '-O3', '-Wall', '-Wstrict-prototypes'])
+                
+            #sysconfig.get_config_var("CFLAGS").split()
+            # includes
+            cmdlist.append("-I" + sysconfig.get_config_var("INCLUDEDIR"))
+            cmdlist.append("-I" + sysconfig.get_config_var("INCLUDEDIR"))
+            cmdlist.append("-I" + sysconfig.get_config_var("INCLUDEPY"))
+            cmdlist.append("-I" + self.python_inc_dir)
+            cmdlist.append("-I" + self.kde_inc_dir)
+            cmdlist.append("-I" + self.qt_inc_dir)
+            cmdlist.append("-I.")
+            # input
+            cmdlist.append(cppfile)
+            # output
+            outputfile = os.path.join(self.build_dir,libraryname+'.lo')
+            cmdlist.append("-o")
+            cmdlist.append(outputfile)
+            spawn(cmdlist) # Execute!!!
+            print
+            
+            # Link the resulting object file to create a shared library.
+            cmdlist = ['libtool']
+            cmdlist.append("--mode=link")
+            
+            # Grab the linker command name
+            cmdlist.append(sysconfig.get_config_var("LDSHARED").split()[0])
+            # link_flags
+            cmdlist.append("-module")
+            cmdlist.append("-export-dynamic")
+            # object
+            cmdlist.append(outputfile)
+            cmdlist.append("-rpath"); cmdlist.append(self.kde_kioslave_lib_dir)
+            cmdlist.append("-o"); cmdlist.append(os.path.join(self.build_dir,stub_la_name))
+            # Link libs
+            linklist = []
+            linklist.append("-lpython%s" % python_version)
+            linklist.extend(sysconfig.get_config_var("LIBS").split())
+            
+            # FIXME I doubt half of these libraries need to be here.
+            linklist.append(self.sip_dir+"/sip.so")
+            # PyQt libs
+            linklist.append(self.pyqt_dir+"/qt.so")
+            # PyKDE libs
+            linklist.append(self.pykde_dir+"/kdecore.so")
+            
+#            linklist.append("-L"+self.sip_dir); linklist.append("-lsip")
+#            # PyQt libs
+#            linklist.append("-L"+self.pyqt_dir); linklist.append("-lqtcmodule")
+#            # PyKDE libs
+#            linklist.append("-L"+self.pykde_dir); linklist.append("-lkdecorecmodule"); linklist.append("-lkdeuicmodule")
+            
+            linklist.append("-L"+self.kde_lib_dir); linklist.append("-lkdecore"); linklist.append("-lpythonize")
+            linklist.append("-L"+self.qt_lib_dir); linklist.append("-lqt-mt")
+            linklist.append("-lm")
+            linklist.append("-lstdc++")
+            linklist.append("-lc")
+            linklist.append(self.clib)
+
+            linklist.append("-R"); linklist.append(self.python_dir)
+            linklist.append("-R"); linklist.append(self.qt_lib_dir)
+            linklist.append("-R"); linklist.append(self.sip_dir)
+            linklist.append("-R"); linklist.append(self.pyqt_dir)
+            linklist.append("-R"); linklist.append(self.pykde_dir)
+            linklist.append("-R"); linklist.append(self.kde_lib_dir)
+
+            cmdlist.extend(linklist)
+            spawn(cmdlist) # Execute!!
+            print
+            
+    cpptemplate = r"""
+/*
+ * Launch Control Centre modules written in Python using an embedded Python
+ * interpreter.
+ * Based on David Boddie's PyKDE-components.
+ */
+
+#include <stdio.h>
+#include <Python.h>
+#include <kinstance.h>
+#define MODULE_DIR "%(moduledir)s"
+#define MODULE_NAME "%(modulename)s"
+#define FACTORY "SlaveFactory"
+
+const char modname[] = MODULE_NAME;
+#define MAIN_METHOD "dispatchLoop"
+
+FILE *d = NULL;
+
+PyObject* call_function(PyObject *callable, PyObject *args) {
+    PyObject *result, *pArgs;
+    
+    if (callable == NULL) {
+        printf(MODULE_NAME " kioslave error: callable == NULL in call_function\n");
+        return NULL;
+    }
+    
+    if (PyCallable_Check(callable)) {
+        if(args == NULL) {
+            pArgs = PyTuple_New(0);
+        } else {
+            pArgs = args;
+        }
+        result = PyObject_CallObject(callable, pArgs);
+        
+        /* If the arguments were created is this function then decrease
+           their reference count. */
+        if(args == NULL) {
+            Py_XDECREF(pArgs);
+            /* pDict and pFunc are borrowed and must not be Py_DECREF-ed */
+        }
+        
+        if(result == NULL) {
+            PyErr_Print();
+            PyErr_Clear();
+        }
+    }
+   
+    return result;
+}
+
+extern "C" {
+    int kdemain( int argc, char **argv) {
+        PyObject *pModule, *pName, *pDict;
+        KInstance slave(MODULE_NAME);
+        
+        Py_SetProgramName(argv[0]);
+        Py_Initialize();
+        //PyEval_InitThreads();
+        PySys_SetArgv(1, argv);
+        
+        PyRun_SimpleString("import sys\n");
+        PyRun_SimpleString("sys.path.append('"MODULE_DIR"')\n");
+        
+        pName = PyString_FromString(modname);
+        pModule = PyImport_Import(pName);
+        
+        Py_XDECREF(pName);
+        
+        if(pModule == NULL) {
+            printf(MODULE_NAME " kioslave error: pModule == NULL\n");
+            return 1;
+        } else {
+            PyObject *pClass, *pMethod, *pArgs, *pArg1, *pArg2, *pInstance;
+            int i;
+            
+            pDict = PyModule_GetDict(pModule);
+            /* pDict is a borrowed reference */
+            
+            pClass = PyDict_GetItemString(pDict, FACTORY);
+            
+            if(pClass == NULL) {
+                printf(MODULE_NAME " kioslave error: pClass == NULL\n");
+                return 1;
+            } else {
+                pArgs = PyTuple_New(2);
+                
+                pArg1 = PyString_FromString(argv[2]);
+                pArg2 = PyString_FromString(argv[3]);
+                
+                PyTuple_SetItem(pArgs, 0, pArg1);
+                PyTuple_SetItem(pArgs, 1, pArg2);
+                
+                call_function(pClass, pArgs);
+                
+                /* Some time later... */
+                Py_XDECREF(pClass);
+                Py_XDECREF(pArgs);
+            }
+            
+            Py_XDECREF(pModule);
+        }
+        
+        Py_Finalize();
+        return 0;
+    }
+}
+"""
+###########################################################################
+class InstallKioslave(Command):
+    description = "Install Kioslave files"
+    
+    user_options = [
+        ('install-dir=', 'd', "base directory for installing kioslave module files"),
+        ('install-cmd=', None, "Command to use to install the files"),
+        ('install-protocol-dir=',None,"directory for kioslave protocol files"),
+        ('build-dir=','b', "build directory (where to install from)"),
+        ('root=', None, "install everything relative to this alternate root directory"),
+        ('force', 'f', "force installation (overwrite existing files)"),
+        ('skip-build', None, "skip the build steps"),
+        ]
+    boolean_options = ['force', 'skip-build']
+    
+    def initialize_options(self):
+        self.build_dir = None
+        self.install_dir = None
+        self.install_cmd = None
+        self.install_protocol_dir = None
+        self.outfiles = []
+        self.root = None
+        self.force = 0
+        self.warn_dir = 1
+        self.skip_build = None
+
+    def finalize_options(self):
+        own_install_dir = self.install_dir is not None
+        own_install_protocol_dir = self.install_protocol_dir is not None
+        
+        self.set_undefined_options('install',
+                                    ('build_base', 'build_dir'),
+                                    ('install_kioslave', 'install_dir'),
+                                    ('root', 'root'),
+                                    ('force', 'force'),
+                                    ('skip_build', 'skip_build'),
+                                    ('install_cmd', 'install_cmd'),
+                                    ('install_protocol','install_protocol_dir')
+                                    )
+
+        if own_install_dir and self.root is not None:
+            self.install_dir = change_root(self.root,self.installdir)
+        
+        if own_install_protocol_dir and self.root is not None:
+            self.install_protocol_dir = change_root(self.root,self.install_protocol_dir)
+        
+    def get_command_name(self):
+        return 'install_kioslave'
+        
+    def run(self):
+        if not self.skip_build:
+            self.run_command('build_kioslave')
+    
+        self.announce("Installing Kioslave files...")
+
+        for moduletuple in self.distribution.kioslaves:
+            self.announce("Building Kioslave module from protocol file %s." % moduletuple[0])
+            
+            protocolfilename = moduletuple[0]
+            
+            # Read the protocol file
+            libraryname = None
+            try:
+                fhandle = open(protocolfilename,'r')
+                for line in fhandle.readlines():
+                    parts = line.strip().split('=')
+                    try:
+                        if parts[0]=="exec":
+                            libraryname = parts[1]
+                    except IndexError:
+                        pass
+                fhandle.close()
+            except IOError:
+                raise SystemExit, "Failed to find kioslave protocol file: %s" % moduletuple[0]
+                
+            if libraryname is None:
+                raise SystemExit, "Failed to find library name (Was there a exec entry in the protocol file?)"
+                
+            self.outfiles.extend(self.mkpath(self.install_protocol_dir))
+            protocolfile_dest = os.path.join(self.install_protocol_dir,os.path.basename(protocolfilename))
+            self.copy_file(protocolfilename, protocolfile_dest)
+            
+            stub_la_name = libraryname+'.la'
+    
+            self.outfiles.extend(self.mkpath(self.install_dir))
+            
+            # Install the library.
+            cmdlist = ['libtool']
+            cmdlist.append("--mode=install")
+            cmdlist.append(self.install_cmd)
+            cmdlist.append("-c")
+            cmdlist.append(os.path.join(self.build_dir,stub_la_name))
+            cmdlist.append(os.path.join(self.install_dir,stub_la_name))
+            spawn(cmdlist) # Execute!!
+            print
+
+            self.outfiles = [os.path.join(self.install_dir,os.path.basename(file)) for file in glob.glob(os.path.join(self.build_dir,'.libs',libraryname+'*'))]
+            self.outfiles.append(protocolfile_dest)
+            
+        self.announce("Done installing Kioslave files.")
+        
+    def get_outputs(self):
+        return self.outfiles or []
+
+    def mkpath(self, name, mode=0777):
+        return dir_util.mkpath(name, mode, dry_run=self.dry_run)
+ 
 ###########################################################################
 class CheckPyQt(Command):
     description = "Checks for the presence of a working PyQt installation"
